-- Sample command line to invoke an interactive interpreter with this
-- help code loaded
-- lua -v "help.lua" -i

-- This is the table that holds the help information. It starts out
-- empty, but the loadHelp () function pulls in help for a particular
-- module and inserts it into this table.
local helpDocs = {}

-- Given a table, this returns an array of all of the keys, sorted, and the length
-- of the longest key.
local collectKeys = function (table)
	local keyList = {}
	local keyCount = 0
	local largest = -1

	-- Go through all keys and add them into the array
	for key, value in table do
		-- Add in
		keyCount = keyCount + 1
		keyList[keyCount] = key

		-- Check maximum length
		if strlen (key) > largest then largest = strlen (key) end
	end

	-- Sort it
	sort (keyList)

	-- Return
	return keyList, largest
end


-- Tells you what modules have help available
local availableModules = function ()
	-- Get the list of modules
	local moduleList,longest = %collectKeys (%helpDocs)

	-- Header
	write ("Modules with help loaded\n")
	write ("------------------------\n\n")

	-- Dump All modules with their synopsis
	for index, name in moduleList do
		local line = format ("%" .. longest .. "s: %s\n" , name , %helpDocs[name]._synopsis)
		write (line)
	end

	write ("\n\n")
end

-- Dump the help text out to screen, stopping to page every linesPerPage
-- lines. On the first loop through, we assume that half of the space
-- per page has already been used. This is sloppy. It would be best if the
-- caller told us exactly how many, but then we'd need to keep count. This
-- should be good enough for now.
local dumpFullHelp = function (linesPerPage,  helpText)
	-- Create a tally for lines printed. We initialize this to be half of the
	-- height we think the terminal is. Once we "fill" the terminal, this will
	-- start at 0 like it should.
	local linesPrinted = linesPerPage / 2

	-- knock linesPerPage down by 1, because we're going to use a line for a 
	-- "-- More --" prompt, and we don't want to scroll the help text off the 
	-- screen when we do.
	linesPerPage = linesPerPage - 1

	-- Split the help text into lines. This requires that every line end in
	-- a newline, but the verify.lua script ensures this.
	local lines = {}
	gsub (helpText , "(.-)".."\n" , function (x) tinsert (%lines,x) end)

	-- Loop printing every line. Whenever we print enough lines, we will stop
	-- and wait for input before continuing.
	for printIndex = 1, getn (lines) do
		-- Print this line and count it.
		write (lines[printIndex] , "\n")
		linesPrinted = linesPrinted + 1

		-- If we printed enough lines, pause.
		if linesPrinted == linesPerPage then
			linesPrinted = 0
			write ("-- Enter for more --")
			read (_STDIN)
		end
	end
	
	write ("\n\n")
end

-- Output help for a record that is a function.
local functionHelper = function (func , moduleName , fullHelp)
	local signature

	-- Create a signature that is the function's name, and it's arguments.
	signature = func._name

	-- If it has arguments, then add those in.
	if func._args ~= nil then
		-- If the first character in the args string is a {. then we assume
		-- that this function is taking a table, so don't wrap the args in
		-- ().
		if strbyte ("{") == strbyte (func._args , 1) then
			signature = signature .. " " .. func._args 
		else
			signature = signature .. " (" .. func._args .. ")"
		end
	else
		signature = signature .. " ()"
	end

	-- Now, we can write stuff out
	local name = func._name .. " (" .. moduleName .. ")"
	write (name , "\n")
	write (strrep ("-" , strlen (name)) , "\n\n")

	write (func._synopsis , "\n\n")

	-- Write out potentially two signatures, one for normal returns, and one
	-- for error returns
	if func._returns then
		write ("Normal:\n\t" , func._returns , " = " , signature , "\n") 
	else
		write ("Normal:\n\t"  , signature , "\n") 
	end

	if func._errorreturns then
		write ("Error:\n\t" , func._errorreturns , " = " , signature , "\n") 
	end

	write ("\n")

	-- If we should display full help, and full help is available, then do it.
	if fullHelp and func._description then %dumpFullHelp (20 , func._description) end
end


-- Output help for a record that is a variable.
local variableHelper = function (var , moduleName , fullHelp)
	-- Write stuff out
	local name = var._name .. " (" .. moduleName .. ")"
	write (name , "\n")
	write (strrep ("-" , strlen (name)) , "\n\n")

	write (var._synopsis , "\n\n")

	write ("\n")

	-- If we should display full help, and full help is available, then do it.
	if fullHelp and var._description then %dumpFullHelp (20 , var._description) end
end


-- A helper that will output the contents of the given array,
-- adding each item to width (left justified).
local printArray = function (array , width)
	-- Determine number of items per line, assuming that we want
	-- to constrain ourself to 75 columns wide.
	local itemsPerLine = floor (75 / width)
	local formatStr = "%-" .. width .. "s"
	local output=0

	for index = 1, getn (array) do
		write  (format (formatStr , array[index]))

		output = output + 1
		if output == itemsPerLine then
			output = 0
			write ("\n")
		end
	end
end


-- A helper that takes a table, collects all of the keys,
-- and then displays them as many to a line as possible in
-- a nice way.
local printModuleContent = function (content , heading)
	-- Collect information about all items
	local itemList, longestItem = %collectKeys (content)

	-- Only proceed if there are some
	if getn (itemList) ~= 0 then
		-- Pad the longest item by 4, to make a bit of a buffer.	
		-- Once this is done, we calculate how many items
		-- we can fit into a 75 column area.
		longestItem = longestItem + 4

		write ("\n" , heading , "\n" , strrep ("-" , strlen (heading)) , "\n")
		%printArray (itemList , longestItem)
		write ("\n\n")
	end
end


-- Output help for a record that is a module.
local moduleHelper = function (module , fullHelp)
	-- Write stuff out
	write (module._name , "\n")
	write (strrep ("-" , strlen (module._name)) , "\n\n")

	write (module._synopsis , "\n\n")

	-- Display variables and functions in this module
	%printModuleContent (module._variables , "Variables")
	%printModuleContent (module._functions , "Functions")

	-- If we should display full help, and full help is available, then do it.
	if fullHelp and module._description then %dumpFullHelp (20 , module._description) end

	write ("\n")
end

-- This is called when there is more than one help item for a
-- given help string. It lists all of the matches, so that
-- a more specific help item can be given. itemList is a table,
-- key = moduleName, value = matchedItem, and moduleName is either
-- a module name that matched the search query, or nil.
local displayMatchList = function (moduleName , itemList)
	-- Output some stuff
	write ("Matching Items\n--------------\n\n")

	if moduleName ~= nil then
		write ("module." , moduleName , " (module)\n")
	end

	for module, item in itemList do
		if strfind (item._name , "." , 1 , 1) == nil then
			write (module , "." , item._name , " (" , item._type , ")\n")
		else
			write (item._name , " (" , item._type , ")\n")
		end
	end

	write ("\n")
end

-- Pull in help for a particular module. Generates a warning if the
-- help file can't be loaded, unless noWarn is not nil. Help files have 
-- the name of a particular module and an extension of .hlp. 
function loadHelp (moduleName , noWarn)
	-- Attempt to load in the help for the given module
	local helpTable = dofile (moduleName .. ".hlp")

	-- If it didn't work, and we're supposed to generate a warning, then
	-- do so.
	if helpTable == nil and noWarn == nil then
		write (_STDERR , "Warning: Unable to load help for module '" .. moduleName .. "'.");
	end

	-- Insert this help into the table at the given index. If the load failed,
	-- this wastes a bit of time but otherwise does nothing.
	%helpDocs[moduleName] = helpTable
end

-- Display all help, including the full help
function fullHelp (helpStr)
	help (helpStr , 1)
end

-- Get help for a module or function. fullHelp is optional. non-nil means to
-- display the full help and not just the synopsis.
function help (helpStr , fullHelp)
	write ("\n")

	-- If no help string is given, then print modules that have help avilable
    -- for them.
	if helpStr == nil or helpStr == "" then
		%availableModules ()
		return
	end

	-- Make a table that will store help matches for functions and variables, and
	-- a variable to hold the name of a module, if one matches. We know, because of
	-- the layout of the help, that there can be only one match per module, and all
	-- modules have a unique name. But we need to guard against a module having the
	-- same name as a function or variable, since modules are just logical groupings,
	-- but the contents of the modules are in the global environment of the interpreter.
	local helpMatches = {}
	local moduleMatches = nil

	-- If the help string has a "." in it, then we assume it is in the format
	-- module.something . In which case, check explicitely for "something" in
	-- "module". If not, then we will scan help for every module.
	--
	-- Either way, matches are stored in the helpMatches table, with a key of
	-- the module name, and a value of the item. It can be a variable or a
	-- function, since the table for each item carries it's own type.
	local start,stop = strfind (helpStr , "%.")
	if start ~= nil then
		-- There is a "." in the string, so split it apart into two pieces. The first part is
		-- going to be a module name, and the second part is going to be the name of a function
		-- or variable inside that module.
		local start,stop,module,item = strfind (helpStr , "([A-Za-z0-9_-]*)%.([A-Za-z0-9_]*)")

		-- Could we not split? If not, then bail.
		if module == nil or item == nil then
			write ("Invalid help string '" , helpStr , "'\n\n")
			return
		end

		-- Search first in variables, then in functions. There can't be two items with the same
		-- name, so order isn't that important.
		if %helpDocs[module] then
			helpMatches[module] = %helpDocs[module]._variables[item] or %helpDocs[module]._functions[item]
		end

		-- If the module name is module, then set moduleMatches to be sub-string. This allows for things like
		-- 'help "module.help"', where help is both a function and a module.
		if module == "module" then moduleMatches = item end
	else
		-- There is no ".", so do a search for all matches. This means we have to scan over all
		-- modules, adding in anything we find. There can only be one match per module, because of the
		-- way our tables are laid out.
		for module, content in %helpDocs do
			-- Search first in variables, then in functions. There can't be two items with the same
			-- name, so order isn't that important.
			helpMatches[module] = %helpDocs[module]._variables[helpStr] or %helpDocs[module]._functions[helpStr]

			-- If the name of this module matches the passed in help string, add it to the moduleMatches list.
			if module == helpStr then moduleMatches = helpStr end
		end
	end

	-- Now, helpMatches is a complete list of everything that matches in all currently available help file, separted
	-- out by module, since we know that the same identifier can't be contained in the same module. Collect the keys
	-- and their length, so we can decide what to do.
	local itemList, longestItem = %collectKeys (helpMatches)

	-- No matches in the item list, and no matching module name, means no help.
	if getn (itemList) == 0 and moduleMatches == nil then
		-- If we get here, no match
		write ("No help available for '" , helpStr , "'\n\n")
		return
	end

	-- If the module name matches, but there's no matches in the item list, then the help string given matches a
	-- module, so output help for it and then bail.
	if moduleMatches ~= nil and getn (itemList) == 0 then
		%moduleHelper (%helpDocs[moduleMatches] , fullHelp)
		return
	end

	-- If there is one item in the helpMatches list, and no match on module name, then the help string uniquely
	-- identified a single item, so display it's help and then leave.
	if getn (itemList) == 1 and moduleMatches == nil then
		-- Get the item we're getting help on
		local moduleName = itemList[1]
		local helpItem = helpMatches[moduleName]

		-- How to display the help depends on what it is
		if helpItem._type == "function" then
			%functionHelper (helpItem , moduleName , fullHelp)
		else
			%variableHelper (helpItem , moduleName , fullHelp)
		end

		return
	end

	-- If we get here, then there is more than one help match and/or more than one module name match. Output a list
	-- of everything that matches, and let the user invoke us again with a more specific name.
	%displayMatchList (moduleMatches , helpMatches)
end


-- Load the help for all of the modules that come standard with the distribution
loadHelp ("core")
loadHelp ("system")
loadHelp ("string")
loadHelp ("io")
loadHelp ("math")
loadHelp ("debug")
loadHelp ("bitlib")
loadHelp ("regex")
loadHelp ("help")
loadHelp ("extended")
loadHelp ("Code")
loadHelp ("FTP")
loadHelp ("HTTP")
loadHelp ("SMTP")
loadHelp ("URL")
loadHelp ("cl-compat")
loadHelp ("luasocket")

